require 'json'
require 'socket'
require 'timeout'
require 'webrick/httpserver'
require 'webrick/https'
require 'uu/os/lang/json_patch'
require 'uu/os/util/url_converter'
require 'uu/os/http/multipart_parser'

# @private
module WEBrick # :nodoc:

  # Extend request class to provide additional ways
  # to process request. Also disables default parsing
  # of multipart requests into query parameters.
  # @private
  class HTTPRequest # :nodoc:
    include UU::OS::HTTP::MultipartParser

    # Returns parsed multipart body, nil if request is not multipart.
    # Do not use for large data, whole request is processed in memory.
    # Method can be invoked only if body was not already consumed in
    # streamed way (neither method {#body_stream} was invoked nor
    # +binary_name+ parameter was given).
    # Unlike default multipart parsing, this method is able to handle
    # header which contains more values than "boundary" alone. And also
    # allows direct parsing of large binary data without using memory
    # or tempfile.
    # @param [Symbol, String] binary_name Name of attribute containing large
    #    binary data. In case attribute with this name is found, its content
    #    is returned as stream and rest of multipart body is dumped (binary
    #    attribute should be always the last item of multipart body).
    # @return [Hash, nil] Parsed body or nil
    def body_multipart(binary_name = nil)
      return @body_multipart if @body_multipart
      if header['content-type'][0] =~ /multipart\/form-data;.*boundary=([^;]+)/
        @body_multipart = {}
        if binary_name.nil?
          parsed = parse_multipart(body, $1)
        else
          parsed = parse_multipart(body_stream, $1, binary_name)
        end
        parsed.each do |item|
          name = item[:name] || item[:filename]
          @body_multipart[name] = item[:body]
        end
      end
      return @body_multipart
    end

    # Returns body as stream. Method can be invoked only if body
    # was not already consumed in streamed way.
    # @return [IO] Body as stream
    def body_stream()
      return StringIO.new(@body) if @body_processed && !@body_destroyed
      raise 'Body was already consumed.' if @body_destroyed
      @body_processed = true
      cont, buff = IO.pipe
      @reader = Thread.new {
        body do |chunk|
          @body_destroyed = true
          buff.write(chunk)
          buff.flush
        end
        buff.close
      }
      return cont
    end

    alias_method :body0, :body
    private :body0

    def body(&block)
      if @body_processed
        if !@body_destroyed
          return @body if block.nil?
          @body.scan(/.{1,4096}/).each do |chunk|
            block.call(chunk)
          end
        else
          raise 'Body was already consumed.'
        end
      end
      @body_processed = true
      @body_destroyed = true if !block.nil?
      body0(&block)
    end

    def parse_query()
      # WEBrick by default parses query parameters only for GET and HEAD
      # In case of POST, as query parameters is parsed content of body,
      # but only in case we have "form" content type. But this is dangerous
      # because of file upload (no streaming).
      @query = HTTPUtils::parse_query(@query_string)
    rescue => e
      raise HTTPStatus::BadRequest, e.message
    end

    def fixup()
      # Need to be patched because of #body_stream method.
      begin
        # Must wait until reading of streamed body is finished.
        @reader.join if !@reader.nil?
        body0 {|chunk|}   # read remaining body
      rescue HTTPStatus::Error => ex
        @logger.error("HTTPRequest#fixup: #{ex.class} occurred.")
        @keep_alive = false
      rescue => ex
        @logger.error(ex)
        @keep_alive = false
      end
    end

  end

  # WEBrick implementation in JRuby has bug (which is fixed in
  # native Ruby) causing concurrent exception under heavy load
  # with multiple running threads.
  # https://www.ruby-forum.com/topic/4405020
  # @private
  module Utils # :nodoc:
    # @private
    class TimeoutHandler # :nodoc:
      def initialize
        @timeout_info = Hash.new
        Thread.start{
          while true
            now = Time.now
            # Original implementation contains
            #   @timeout_info.each{...
            # which causes problems because hash
            # cannot be extended wile iterating
            # over it.
            @timeout_info.keys.each{|thread|
              ary = @timeout_info[thread]
              next unless ary
              ary.dup.each{|info|
                time, exception = *info
                time ||= Time.now
                interrupt(thread, info.object_id, exception) if time < now
              }
            }
            sleep 0.5
          end
        }
      end
    end

  end

end

module UU
  module OS
    module Util

      # Simple HTTP server based on WEBrick to be used mainly for testing purposes. Additionally
      # when this class is required, {WEBrick::HTTPRequest WEBrick::HTTPRequest} object is extended
      # with methods +#body_stream+ and +#body_multipart+ to ease processing of incoming requests.
      # @example Basic usage:
      #  require 'uu/os/con/util/embedded_http_server'
      #
      #  server = UU::OS::CON::Util::EmbeddedHTTPServer.new
      #  server.register_servlet(My::Module::MyServlet, '/path/to/my/servlet')
      #  server.start
      #  ...
      #  server.shutdown
      class EmbeddedHTTPServer
        include UU::OS::Util::URLConverter

        # Default host
        # @private
        DEFAULT_HOST = 'localhost' # :nodoc:
        private_constant :DEFAULT_HOST

        # Server host.
        # @return [String] Server host.
        attr_reader :host

        # Server port.
        # @return [Fixnum] Server port.
        attr_reader :port

        # Server context path.
        # @return [String] Server context path.
        attr_reader :context

        # Server base URL (without context path).
        # @return [String] Server base URL.
        attr_reader :base_url

        # Creates new instance of server.
        # @param [Hash] options Server options:
        # @option options [String] :host Server host (defaults to +localhost+).
        # @option options [Fixnum] :port Server port (defaults to first available unused port).
        # @option options [String] :context Server context path (defaults to empty path).
        # @option options [String, IO] :log_file Path to server log file or existing IO (defaults to +$stderr+).
        # @option options [Fixnum] :log_level Log level (possible values 0=OFF, 1=FATAL, 2=ERROR, 3=WARN, 4=INFO, 5=DEBUG; defaults to 2=ERROR).
        # @option options [String, IO] :access_log_file Path to server access log file or existing IO (defaults to +$stderr+).
        # @option options [TrueClass, FalseClass] :access_log Flag if access log is enabled (defaults to +false+).
        # @option options [TrueClass, FalseClass] :use_ssl Flag if server should use SSL (default to +false+). If options :ssl_cert and
        #    :ssl_pkey are not provided, server uses self signed certificate generated with each start.
        # @option options [String, IO] :ssl_cert Path to SSL certificate or existing IO (defaults to +nil+)
        # @option options [String, IO] :ssl_pkey Path to SSL certificate private key or existing IO (defaults to +nil+)
        def initialize(options = {})
          @host = options[:host] || DEFAULT_HOST
          @port = options[:port] || empty_port
          @context = options[:context] || ''
          @base_url = "#{options[:use_ssl] ? "https" : "http"}://#{@host}:#{@port}"
          config = {BindAddress: @host, Port: @port}
          config[:Logger] = WEBrick::Log.new(options[:log_file], (options[:log_level] || 2))
          if options[:access_log_file]
            log = options[:access_log_file]
            log = File.open(log, 'a') if log.kind_of?String
            config[:AccessLog] = [[log, AccessLog::COMMON_LOG_FORMAT], [log, AccessLog::REFERER_LOG_FORMAT]]
          elsif !options[:access_log]
            config[:AccessLog] = []
          end
          if options[:use_ssl]
            config[:SSLEnable] = true
            cert = options[:ssl_cert]
            pkey = options[:ssl_pkey]
            if cert.nil? || pkey.nil?
              config[:SSLCertName] = [['CN', @host]]
            else
              cert = File.open(cert, 'rb') if cert.kind_of?String
              pkey = File.open(pkey, 'rb') if pkey.kind_of?String
              config[:SSLCertificate] = OpenSSL::X509::Certificate.new(cert.read) if !cert.nil?
              config[:SSLPrivateKey] = OpenSSL::PKey::RSA.new(pkey.read) if !pkey.nil?
            end
          end
          @server = WEBrick::HTTPServer.new(config)
        end

        # Registers servlet to be invoked for given path.
        # @param [WEBrick::HTTPServlet::AbstractServlet] servlet_class Any class inherited from +AbstractServlet+.
        # @param [Array<String>] path Servlet path (can be given as separate path elements). If no path is given,
        #    servlet is registered to root of context.
        # @return [void]
        def register_servlet(servlet_class, *path)
          if path.size > 0
            deploy_path = concat_path('', @context, *path)
          else
            deploy_path = concat_path('', @context, '/')
          end
          @server.mount(deploy_path, servlet_class)
        end

        # Starts server.
        # @return [void]
        def start
          @thread = Thread.new {
            Thread.current.abort_on_exception = true
            @server.start
          }
          while @server.status != :Running
            Thread.pass
            unless @thread.alive?
              @thread.join
              raise
            end
          end
          return
        end

        # Stops server.
        # @return [void]
        def stop
          @server.stop
          return
        end

        # Stops server and releases all listening sockets. Server
        # stopped via this method cannot be restarted (new server
        # instance must be created).
        # @return [void]
        def shutdown
          @server.shutdown
          return
        end

        # Returns server status.
        # @return [Symbol] Server status.
        def status
          @server.status
        end

        # Override inspect to hide internal structures.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)} @base_url=\"#{base_url}\", @status=#{status}>"
        end

        private

        # Returns first available unused port.
        # @return [Fixnum] Unused port.
        # @private
        def empty_port # :nodoc:
          srv = TCPServer.new(@host, 0)
          return srv.addr[1]
        ensure
          srv.close if !srv.nil?
        end

      end

    end
  end
end
